// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "bhv_basic_move.h"

#include "bhv_basic_tackle.h"

#include <rcsc/action/basic_actions.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/body_intercept.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>
#include <rcsc/player/intercept_table.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>

#include <map>
#include <list>
#include <set>


/*-------------------------------------------------------------------*/
/*!
  execute action
*/
bool
Bhv_BasicMove::execute( rcsc::PlayerAgent * agent )
{
    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        "%s:%d: Bhv_BasicMove"
                        ,__FILE__, __LINE__ );

    // tackle

    if ( Bhv_BasicTackle( 0.9, 80.0 ).execute( agent ) )
    {
        return true;
    }

    const rcsc::WorldModel & wm = agent->world();
    const rcsc::PlayerPtrCont & mates = wm.teammatesFromSelf();
    const rcsc::PlayerPtrCont & opps  = wm.opponentsFromSelf();

    // check ball owner

    int self_min = wm.interceptTable()->selfReachCycle();
    int mate_min = wm.interceptTable()->teammateReachCycle();
    int opp_min = wm.interceptTable()->opponentReachCycle();
    static int runningCount = 0;
    static rcsc::Vector2D runningTarget;
    // check ball owner

    int  smax = 3;
    int smmax = 3;
    int somax = 4;

    if (wm.self().unum() == 6 || 
	wm.self().unum() == 7) {
      smax = 7;
      somax = 11;
    }

    if ( ! wm.existKickableTeammate()
         && ( self_min <= smax//3
              || ( self_min < mate_min + smmax
                   && self_min < opp_min + somax) //4 )
              )
         )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: intercept"
                            ,__FILE__, __LINE__ );
        rcsc::Body_Intercept().execute( agent );
        if ( wm.ball().distFromSelf()
             < rcsc::ServerParam::i().visibleDistance() )
        {
            agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );;
        }
        else
        {
            agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        }
        return true;
    }

    if (wm.self().unum() > 5) {
    if ( (! wm.existKickableTeammate()) &&
	 runningCount >  0) {
      runningCount --;
      rcsc::Body_GoToPoint(runningTarget, 1.0, rcsc::ServerParam::i().maxPower(),10,false,true ).execute(agent);
      agent->setNeckAction( new rcsc::Neck_TurnToBall() );
      std::cout << wm.time() << " running"
		<< wm.self().unum() << runningTarget << std::endl;
      return true;
    }

    if ( wm.existKickableTeammate() &&
	 (wm.self().pos().x < wm.offsideLineX()) &&
	 (wm.offsideLineX() - wm.self().pos().x < 5) ) {
      // ボールを蹴れそうなやつがおって自分にスルーパスを
      // 出すような気がしたらとりあえず５ステップ走っておく
      const rcsc::PlayerPtrCont::const_iterator mates_end = mates.end();
      rcsc::PlayerPtrCont::const_iterator it;
      for ( it = mates.begin();  it != mates_end; ++it ) {
	if ((*it)->isKickable()) break;
      }
      rcsc::PlayerPtrCont::const_iterator kickable = it;
      if (it != mates_end && (*it)->pos().x >= 0.0) {
	const rcsc::PlayerPtrCont::const_iterator opps_end = opps.end();
	std::set< double > stmp;
	const double angleRange    = 70;
	const double lowerBoundGap = 20;
	for ( it = opps.begin();  it != opps_end; ++it ) {
	  if ((*it)->goalie()) continue;
	  rcsc::Vector2D re_v = (*it)->pos() - (*kickable)->pos();
	  double angle = re_v.th().degree();
	  if (angle > angleRange) continue;
	  if (angle < -angleRange) continue;
	  stmp.insert(angle);
	}
	std::map<double, double> mtmp;
	if (stmp.size() > 0)  {
	  double prev = -angleRange;
	  std::set< double >::iterator is;
	  for (is = stmp.begin(); is != stmp.end(); ++is) {
	    if ((*is) - prev > lowerBoundGap) {
	      mtmp[angleRange - prev] = ((*is) + prev)/2 ;
	    }
	    prev = (*is);
	  }
	  if (angleRange - prev > lowerBoundGap) { 	    
	    mtmp[angleRange - prev] = (angleRange + prev)/2 ;
	  }
	}
	std::map<double, double>::reverse_iterator im;
	for (im = mtmp.rbegin(); im != mtmp.rend(); ++im) {
	  rcsc::AngleDeg a1 = 0;
	  rcsc::Line2D lc(wm.self().pos(), a1);
	  rcsc::AngleDeg a2 = im->second;
	  rcsc::Line2D ls((*kickable)->pos(), a2);
	  rcsc::Vector2D vi = ls.intersection(lc);
	  if (vi.x > wm.self().pos().x + 5 &&  vi.x < 50)   {
	    runningCount = 4;
	    runningTarget = vi;
	    rcsc::Body_GoToPoint(runningTarget, 1.0, rcsc::ServerParam::i().maxPower(),10,false,true ).execute(agent);
	    agent->setNeckAction( new rcsc::Neck_TurnToBall() );
	    return true;
	  }
	}

      }
      
    }




    if(wm.existKickableTeammate() && wm.self().pos().x < 36.0 &&
	wm.self().pos().x > wm.offsideLineX()-10 ) {
     const rcsc::PlayerPtrCont & opps  = wm.opponentsFromSelf();
     const rcsc::PlayerPtrCont::const_iterator opps_end = opps.end();
     rcsc::PlayerPtrCont::const_iterator it;
     double nearest = 10000;
     rcsc::PlayerObject *nearestY = NULL;
     rcsc::PlayerObject *upperSNearestY = NULL;
     rcsc::PlayerObject *lowerSNearestY = NULL;
     for ( it = opps.begin();  it != opps_end; ++it ) {
       if ((*it)->rpos().x < 0) continue;
       if ((*it)->pos().x < wm.offsideLineX() - 5.0) continue;
       if ((*it)->rpos().y < nearest) {
	 nearest = (*it)->rpos().y;
	 if (nearestY && nearestY->pos().y > (*it)->pos().y)
	   upperSNearestY = nearestY;
	 if (nearestY && nearestY->pos().y < (*it)->pos().y)
	   lowerSNearestY = nearestY;
	 nearestY = (*it);
       }
     }
     if (nearestY) {
       double upperBound = (upperSNearestY) ? upperSNearestY->pos().y : 32;
       double lowerBound = (lowerSNearestY) ? lowerSNearestY->pos().y : -32;
       rcsc::Vector2D dash_point;
       if (upperBound - nearestY->pos().y  > lowerBound - nearestY->pos().y ) {
	 dash_point.assign(wm.offsideLineX() - 1.0, 
			   upperBound + nearestY->pos().y / 2.0);
       }
       else {
	 dash_point.assign(wm.offsideLineX() - 1.0, 
			   lowerBound + nearestY->pos().y / 2.0);
       }
       std::cout << wm.time() << wm.self().unum() << " dash " << dash_point << std::endl;
       rcsc::Body_GoToPoint(dash_point, 1.0, rcsc::ServerParam::i().maxPower()/2,10,false,true ).execute(agent);
       agent->setNeckAction( new rcsc::Neck_TurnToBall() );
     }
     else {
       rcsc::Body_TurnToBall().execute( agent );//face to ball position
       agent->setNeckAction( new rcsc::Neck_TurnToBall() );
     }
     return true;
   }

      
    if(wm.existKickableOpponent() && 
       wm.ball().pos().x < 0.0 &&
       wm.self().pos().x < 0.0) {
      const rcsc::PlayerPtrCont & opps  = wm.opponentsFromSelf();
      const rcsc::PlayerPtrCont::const_iterator opps_end = opps.end();
      const rcsc::PlayerPtrCont::const_iterator mates_end = mates.end();
      rcsc::PlayerPtrCont::const_iterator it;
      rcsc::PlayerObject *target = NULL;
      std::map< double, rcsc::PlayerObject * > oppsIN;
      std::map< double, rcsc::PlayerObject * >::iterator im;
      std::list< rcsc::PlayerObject * > matesIN;
      std::list< rcsc::PlayerObject * >::iterator il;
      std::list< rcsc::PlayerObject * >::iterator il2;
      for ( it = opps.begin();  it != opps_end; ++it ) {
	if ((*it)->pos().x < 0) {
	  oppsIN[wm.ball().pos().dist((*it)->pos())] = *it;
	  if ((*it)->isKickable() ) target = *it;
	}
      }

      
      for ( it = mates.begin();  it != mates_end; ++it ) 
	if ((*it)->pos().x < 0) matesIN.push_back(*it);

      if (target && oppsIN.size() < matesIN.size()) {
	for(im = oppsIN.begin(); im != oppsIN.end(); ++im) {
	  il2 = matesIN.end();
	  double min = 10000000;
	  for(il = matesIN.begin(); il != matesIN.end(); ++il) {
	    double dtmp = im->second->pos().dist((*il)->pos());
	    if (min > dtmp) {
	      min = dtmp;
	      il2 = il;
	    }
	  }
	  if (il2 != matesIN.end())
	    matesIN.erase(il2);
	}
	il2 = matesIN.end();
	double min = 10000000;
	for(il = matesIN.begin(); il != matesIN.end(); ++il) {
	  double dtmp = target->pos().dist((*il)->pos());
	  if (min > dtmp) {
	    min = dtmp;
	    il2 = il;
	  }
	  if (il2 != matesIN.end() && (*il2)->unum() == wm.self().unum()) {
	    if ( Bhv_BasicTackle( 0.9, 80.0 ).execute( agent ) ) {
		return true;
	    }
	    rcsc::Body_GoToPoint(target->pos(), 1.0, rcsc::ServerParam::i().maxPower(),10,false,true ).execute(agent);
	    return true;
	  }	    
	}
	  
      }

    }

    }

    // go to home position

    const double dash_power = getDashPower( agent, M_home_pos );

      double dist_thr = wm.ball().distFromSelf() * 0.1;
      if ( dist_thr < 1.0 ) dist_thr = 1.0;

    agent->debugClient().addMessage( " BasicMove%.0f", dash_power );
    agent->debugClient().setTarget( M_home_pos );

      if ( ! rcsc::Body_GoToPoint( M_home_pos, 1.0, dash_power
                                 ).execute( agent ) )
    {
      rcsc::Body_TurnToBall().execute( agent );//face to ball position
	}

    if ( wm.existKickableOpponent()
         && wm.ball().distFromSelf() < 18.0 )
    {
        agent->setNeckAction( new rcsc::Neck_TurnToBall() );
    }
    else
    {
        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
    }

    return true;
}

/*-------------------------------------------------------------------*/
/*!

*/
double
Bhv_BasicMove::getDashPower( const rcsc::PlayerAgent * agent,
                             const rcsc::Vector2D & target_point )
{
    static bool s_recover_mode = false;

    const rcsc::WorldModel & wm = agent->world();
    /*--------------------------------------------------------*/
    // check recover

    if ( wm.self().stamina() < rcsc::ServerParam::i().staminaMax() * 0.5 )
    {
        s_recover_mode = true;
    }
    else if ( wm.self().stamina() > rcsc::ServerParam::i().staminaMax() * 0.7 )
    {
        s_recover_mode = false;
    }

    /*--------------------------------------------------------*/
    double dash_power = 100.0;
    const double my_inc
        = wm.self().playerType().staminaIncMax()
        * wm.self().recovery();

    if ( wm.defenseLineX() > wm.self().pos().x
         && wm.ball().pos().x < wm.defenseLineX() + 20.0 )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: correct DF line. keep max power"
                            ,__FILE__, __LINE__ );
        // keep max power
        dash_power = rcsc::ServerParam::i().maxPower();
    }
    else if ( s_recover_mode )
    {
        dash_power = my_inc - 25.0; // preffered recover value
        if ( dash_power < 0.0 ) dash_power = 0.0;

        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: recovering"
                            ,__FILE__, __LINE__ );
    }
    // exist kickable teammate
    else if ( wm.existKickableTeammate()
              && wm.ball().distFromSelf() < 20.0 )
    {
        dash_power = std::min( my_inc * 1.1,
                               rcsc::ServerParam::i().maxPower() );
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: exist kickable teammate. dash_power= %f"
                            ,__FILE__, __LINE__,
                            dash_power );
    }
    // in offside area
    else if ( wm.self().pos().x > wm.offsideLineX() )
    {
        dash_power = rcsc::ServerParam::i().maxPower();
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: in offside area. dash_power= %f"
                            ,__FILE__, __LINE__,
                            dash_power );
    }
    // normal
    else
    {
        dash_power = std::min( my_inc * 1.7,
                               rcsc::ServerParam::i().maxPower() );
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: normal mode dash_power= %f"
                            ,__FILE__, __LINE__,
                            dash_power );
    }

    return dash_power;
}
